function [cluster_id, order, d, sorted_cluster_id, inlier, n_motions] = cluster_samples(residuum_mat, outl_thresh, n_motions, method)
% 
% [cluster_id, order, d, sorted_cluster_id, inlier, n_motions] = cluster_samples(residuum_mat, outl_thresh, n_motions, method)
% 
% residuum_mat: linear residual with the samples in the colums and observations in
% the rows. Mark unknown data with NaN.
% d is the affinity matrix used for clustering. The affinity matrix hast
% HIGH values on its diagonal!

if nargin < 4
    method = 'kmeans';
end

fprintf('Clustering residuum matrix of size %dx%d into %d classes\n', size(residuum_mat, 1), ...
        size(residuum_mat, 2), n_motions);



%% residuum_mat is a matrix of size MxN (M iterations, N samples)
[M, N] = size(residuum_mat);


%% cluster the observations (transposed: samples, non-transposed: iterations)

switch method
    case 'none'
        D = pdist_jb(residuum_mat, outl_thresh);
        d = 1 - squareform(D);
        cluster_id = round(rand(1, M)) + 1;
        
    case 'kmeans'   
        % d is of size NxN
        tstart=tic;
        D = pdist_jb(residuum_mat, outl_thresh);
        d = 1 - squareform(D);
        time_dist=toc(tstart);
        
        tstart=tic;
        [cluster_id, centers, dist] = kmeans(d, n_motions, 'EmptyAction', 'drop', ...
            'Replicates', 10);
        time_cluster=toc(tstart);
        fprintf('%.1f and %.1f ms for distance computation and clustering (%dx%d)\n', ...
            time_dist*1000, time_cluster*1000, M, N);
        
        if find(isnan(centers))
            fprintf('clustering failed, just splitting into two arbitary parts\n');
            cluster_id = round(rand(1, numel(cluster_id))) + 1;
        end
        
        if 1
            % dist is sum of distances to cluster center
            occurences=hist(cluster_id, 1:n_motions);
            md_cluster=dist'./ occurences; %md is mean distance to cluster center
            mde = md_cluster / N; % mde is mean element-wise distance
            mdestr = sprintf('%f ', mde);
            fprintf('clustered with mdes=[%s], overall mde=%f.\n', mdestr, mean(mde));
        end

    case 'kmeans2'   
        % d is of size NxN
        tstart=tic;
        D = pdist_jb(residuum_mat, outl_thresh);
        d = 1 - squareform(D);
        time_dist=toc(tstart);
        
        tstart=tic;
        [cluster_id, centers, dist] = kmeans2(d, n_motions);
        time_cluster=toc(tstart);
        fprintf('%.1f and %.1f ms for distance computation and clustering (%dx%d)\n', ...
            time_dist*1000, time_cluster*1000, M, N);
        
        if find(isnan(centers))
            fprintf('clustering failed, just splitting into two arbitary parts\n');
            cluster_id = round(rand(1, numel(cluster_id))) + 1;
        end
        
        if 1
            % dist is sum of distances to cluster center
            occurences=hist(cluster_id, 1:n_motions);
            md_cluster=dist'./ occurences; %md is mean distance to cluster center
            mde = md_cluster / N; % mde is mean element-wise distance
            mdestr = sprintf('%f ', mde);
            fprintf('clustered with mdes=[%s], overall mde=%f.\n', mdestr, mean(mde));
        end
        
     case 'spectral'   
        % d is of size NxN
        D = pdist_jb(residuum_mat, outl_thresh);
        d = 1 - squareform(D);
        
        % it would be best if we could use the errorsum value
        sim = d;
        %sim = exp(-d.^2);
        %sim = d.^2;
        %sim = log(d+eps);
        %sim = 1./(1-d-eps);
        
        [diagMat,LMat,X,Y,cluster_id,errorsum]= spectralcluster(sim, n_motions+1, n_motions);

     case 'spectral2'   
        % d is of size NxN
        D = pdist_jb(residuum_mat, outl_thresh);
        d = squareform(D);
        
        % it would be best if we could use the errorsum value
        sim = 1 - d;
        %sim = exp(-5*d.^2); % .1: O (0.465), 5: S (0.481), much worse reuslts
        %sim = exp(d);
        %sim = d.^2;
        %sim = log(d+eps);
        %sim = 1./(1-d-eps);
        
        [diagMat,LMat,X,Y,cluster_id,errorsum]= spectralcluster(sim, ...
                                                  n_motions+1, n_motions, ...
                                                  'kmeans2');
        d = sim; % convenient drawing
        
     case 'normalcut'   
        % d is of size NxN
        D = pdist_jb(residuum_mat, outl_thresh);
        d = 1 - squareform(D);
        
        [cluster_id,cheeger]= spectralclusternormalcut(d);

    case 'meanshift'
        D = pdist_jb(residuum_mat, outl_thresh);
        d = 1 - squareform(D);
        
        %%
        bw = 13.11011
        [cluster_id, m] = meanShift(d, bw);
        max(cluster_id)
        %%
        
    case 'kmeans_old'   
        sampled = ~isnan(residuum_mat);
        inlier = residuum_mat < outl_thresh;
        inlier(~sampled) = false;
        % the jaccard distance is 0 on the diagonal and 1 for unrelated
        % data
        D = pdist(double(inlier'), 'jaccard');
        d = 1 - squareform(D); % 0: worst, 1: best
        
        d(isnan(d)) = 0;
        [cluster_id, centers] = kmeans(d, n_motions, 'EmptyAction', 'drop', ...
                                'Replicates', 3);
        
        if find(isnan(centers))
            fprintf('clustering failed, just splitting into two arbitary parts\n');
            cluster_id = round(rand(1, numel(cluster_id))) + 1;
        end

    case 'jlinkage'     
        sampled = ~isnan(residuum_mat);
        inlier = residuum_mat < outl_thresh;
        inlier(~sampled) = false;

        % exactly the same as above
        D = pDistJaccard(inlier);
        d = 1 - squareform(D);
        
        Z = linkageIntersect(D, inlier');
        cluster_id = cluster(Z, 'cutoff', 1-1/N+eps, 'criterion', 'distance');
        
    otherwise
        fprint('Unknown method: %s\n', method);
end

%% change the clustering order such that big clusters get lower ids
cluster_id = sort_classes(cluster_id);

%%
[sorted_cluster_id,order] = sort(cluster_id);
a=jcorr(cluster_id, d)

sa=sum(sum(a));
sd=sum(diag(a));

q = sd / sa;
fprintf('Class separation factor: %.3f\n', q);