function [idx, order, d, sorted_idx, inlier, n_motions] = ...
    cluster_samples2(residuum_mat, data_mat, outl_thresh, n_motions, method)
% 
% [idx, order, d, sorted_idx, inlier, n_motions] =
% cluster_samples2(residuum_mat, data_mat, outl_thresh, n_motions, method)
% 
% residuum_mat: linear residual with the samples in the colums and observations in
% the rows. Mark unknown data with NaN.

if nargin < 5
    method = 'kmeans';
end

%% residuum_mat is a matrix of size MxN (M iterations, N samples)
[M, N] = size(residuum_mat);


%% cluster the observations (transposed: samples, non-transposed: iterations)

switch method
    case 'none'
        D = pdist_jb(residuum_mat, outl_thresh);
        d = 1 - squareform(D);
        idx = round(rand(1, M)) + 1;
        
    case 'kmeans'   
        % d is of size NxN
        tstart=tic;
        % motion distance
        DM = pdist_jb(residuum_mat, outl_thresh);
        dm = 1 - squareform(DM);
        % spatial distance in second image
        DS = pdist(data_mat(:, 4:5), 'seuclidean');
        ds = exp(- squareform(DS));
        d = dm.*ds;
        time_dist=toc(tstart);
        
        tstart=tic;
        [idx, centers, dist] = kmeans(d, n_motions, 'EmptyAction', 'drop', ...
            'Replicates', 3);
        time_cluster=toc(tstart);
        fprintf('%.1f and %.1f ms for distance computation and clustering (%dx%d)\n', ...
            time_dist*1000, time_cluster*1000, M, N);
        
        if find(isnan(centers))
            fprintf('clustering failed, just splitting into two arbitary parts\n');
            idx = round(rand(1, numel(idx))) + 1;
        end
        
        if 1
            % dist is sum of distances to cluster center
            % dist_i = sum_x_j in c_i ( |x_j - c_i| )
            occurences=hist(idx, 1:n_motions);
            md_cluster=dist'./ occurences; %md is mean distance to cluster center
            mde = md_cluster / N; % mde is mean element-wise distance
            mdestr = sprintf('%f ', mde);
            fprintf('clustered with mdes=[%s], overall mde=%f.\n', mdestr, ...
                    mean(mde));
            
            a=jcorr(idx, d)
        end
        
    case 'meanshift'
        D = pdist_jb(residuum_mat, outl_thresh);
        d = 1 - squareform(D);
        
        %%
        bw = 13.11011
        [idx, m] = meanShift(d, bw);
        max(idx)
        %%
    case 'kmeans_old'   
        sampled = ~isnan(residuum_mat);
        inlier = residuum_mat < outl_thresh;
        inlier(~sampled) = false;
        % the jaccard distance is 0 on the diagonal and 1 for unrelated
        % data
        D = pdist(double(inlier'), 'jaccard');
        d = 1 - squareform(D); % 0: worst, 1: best
        
        d(isnan(d)) = 0;
        [idx, centers] = kmeans(d, n_motions, 'EmptyAction', 'drop', ...
                                'Replicates', 3);
        
        if find(isnan(centers))
            fprintf('clustering failed, just splitting into two arbitary parts\n');
            idx = round(rand(1, numel(idx))) + 1;
        end

    case 'jlinkage'     
        sampled = ~isnan(residuum_mat);
        inlier = residuum_mat < outl_thresh;
        inlier(~sampled) = false;

        % exactly the same as above
        D = pDistJaccard(inlier);
        d = 1 - squareform(D);
        
        Z = linkageIntersect(D, inlier');
        idx = cluster(Z, 'cutoff', 1-1/N+eps, 'criterion', 'distance');
        
        n_motions = max(idx);
        
    otherwise
        disp 'Unknown method: ' method;
end

[sorted_idx,order] = sort(idx);
