/*******************************************************
% Calculates pairwise jaccard distance
%
% Authors: R.Toldo A.Fusiello, department of computer science - University of Verona.
% Reference Paper: R. Toldo, A. Fusiello. Robust Multiple Structures Estimation with J-linkage. Proceeding of the European Conference on Computer Vision, 2008.
% modified version by Ralf Dragon
*******************************************************/

#include "mex.h"
#include <math.h>
#include <string.h>

/* Jaccard distance */
void jacdist(double *x, double thresh, mwSize m, mwSize n, double *d)
{
     // jaccard distance:
     // (#(a|b) - #(a&b)) / #(a|b)
     // n_xor / n_or
     mwIndex i,j,k;
     int   n_xor, n_or;
     double  *XI, *XJ, *XI0;
     bool b1, b2;
	
     XI = x;
     for (i=0; i<m; i++) {
          XI0 = XI;
          XJ = XI+n;
          for (j=i+1; j<m; j++) {
               XI = XI0;
               
               n_xor = 0; //
               n_or = 0;     // 
               for (k=0; k<n; k++, XI++, XJ++) {
                    // if one of both is nan, it the other should not be used
                    if (isnan(*XI) | isnan(*XJ))
                         continue;

                    b1 = *XI < thresh;
                    b2 = *XJ < thresh;
                    
                    n_or += b1|b2;
                    n_xor += b1^b2;
               }
               if (n_or) {
                    *(d++) = (double)n_xor/(double)n_or;
               } else {
                    *(d++) = 1; // only outliers or nans
               }
          }
     }
}


/* the dispatcher function */
void distfun(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
     mwSize numCoords, numPoints;
     double *x, *thresh, *d;

     /*  create a pointer to the input matrix y */
     x = (double *) mxGetPr(prhs[0]);
     thresh = (double *) mxGetPr(prhs[1]);

     /*  get the dimensions of the matrix input y */
     numCoords = mxGetM(prhs[0]);
     numPoints = mxGetN(prhs[0]);

     /* make sure that the distance matrix can be created, then create it.  doing
      * this in double remains exact except in the cases where we error out anyway. */
     double numDists = ((double)numPoints * (double)(numPoints-1)) / 2;
     if (numDists >= (double)MWSIZE_MAX) {
          mexErrMsgIdAndTxt("stats:pdistmex:OutputTooLarge",
                            "Distance matrix has more elements than the maximum allowed size in MATLAB.");
     }
     plhs[0] = mxCreateDoubleMatrix(1, (mwSize)numDists, mxREAL);
    
     /*  create a pointer to a copy of the output matrix */
     d = (double *)mxGetData(plhs[0]);
     jacdist(x, *thresh, numPoints, numCoords, d);
 }

/* the gateway function */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{

  /*  check for proper number of arguments */
    if (nrhs<2) {
        mexErrMsgIdAndTxt("stats:pdistmex:TooFewInputs",
                          "Two input argument required (residual mat, threshold).");
    } else if(nlhs>1) {
        mexErrMsgIdAndTxt("stats:pdistmex:TooManyOutputs",
                          "Too many output arguments.");
    }

    //printf("double: %s", mxIsDouble(prhs[1]) ? "1":"0");

  /* Check the type of the input array */
  /* Currently only works with double or single(float) */
    if (mxIsDouble(prhs[0])) {
        distfun(nlhs, plhs, nrhs, prhs);
    } else {
        mexErrMsgIdAndTxt("stats:pdistmex:BadInputType",
                          "PDISTJaccardMEX only supports real Logicals data.");
    }
}
